package ssw.annotation.annoview;

import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.ui.part.*;
import org.eclipse.ui.texteditor.AbstractTextEditor;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.text.TextSelection;
import org.eclipse.jface.viewers.*;
import org.eclipse.jface.action.*;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.*;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.SWT;
import ssw.annotation.AnnoContentProvider;
import ssw.annotation.AnnotationPlugin;
import ssw.annotation.AnnotationFileHandler.AnnoObject;
import ssw.annotation.AnnotationFileHandler.AnnotationFile;

/*
 * Created on 12.02.2005
 */
/**
 *   Copyright (C) 2005  Stefan Loidl
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *   
 *   eMail: stefanloidl@gmx.net
 *   
 * This view is meant to manage the annotation reviewing cycle. It can be opened using 
 * the Eclipse standard feature to open views, but if an annotation editor is opened 
 * the view is opened anyway. It shows a list of annotations, using the AnnoContentProvider. 
 * The table contains: reviewed (checked), description, file and point modification of the 
 * annotations data. 
 * The data is managed by an underlying AnnotationFile which is used to update changed data too.
 * If modifications are done within the view, they are written to the resource file immediately. 
 * The user is able to modify the checked status to signal that the annotation was reviewed. 
 * Moreover the annotations can be deleted. These actions may be carried out using several 
 * different possibilities. The first possibility is to use the popup menu that is hooked to 
 * the annotation list table or in case of the restore operation a simple double click action 
 * may be used. Another possibility is to use keyboard hotkeys; and last but not least the 
 * icons in the corresponding action bar can be used.
 * The main feature of the view is to restore the highlighted area in the java file that is 
 * referenced by the annotation. This is done using the file, offset and length data from the 
 * annotation item. The offset is the sign count from beginning of the file indicating the 
 * beginning of the highlighted area. In combination with the length the area can be 
 * reconstructed. The reconstruction is shown in the compilation unit editor usually used 
 * for displaying java files. 
 * The view listens for resource change events to update the annotation list on a change event. 
 * For example if manipulations are done within the editor the annotation file is refreshed. 
 * 
 * @version 1.0.0
 * @author Loidl Stefan, 0255886
 */

public class AnnotationView extends ViewPart implements IResourceChangeListener, KeyListener {

	private TableViewer viewer;

	//Actions
	private Action actionRemove;
	private Action actionReviewed;
	private Action doubleClickAction;

	private final String JAVAEDITORID="org.eclipse.jdt.ui.CompilationUnitEditor";
	private final String POPUPMENUID="#AnnoViewPopup";

	//keyboard values
	private final int KEY_ENF=127;
	private final int KEY_R=114;
	private final int KEY_S=115;
	
	//table column Widths
	private int stateWidth=20;
	private int pointWidth=50;
	private int descWidth=800;
	private int fileWidth=250;
	
	private String[] columnLabels=new String[4];

	/**
	 * 
	 * @uml.property name="annoFile"
	 * @uml.associationEnd multiplicity="(0 1)"
	 */
	private AnnotationFile annoFile = null;

	/**
	 * 
	 * @uml.property name="contentProvider"
	 * @uml.associationEnd multiplicity="(0 1)"
	 */
	private AnnoContentProvider contentProvider;

	/**
	 * Constructor.
	 */
	public AnnotationView() {

		columnLabels[0]=AnnotationPlugin.getResourceString("AV_C_REVIEWED");
		columnLabels[1]=AnnotationPlugin.getResourceString("AV_C_DESCRIP");
		columnLabels[2]=AnnotationPlugin.getResourceString("AV_C_POINTS");
		columnLabels[3]=AnnotationPlugin.getResourceString("AV_C_FILE");
		
		ResourcesPlugin.getWorkspace().addResourceChangeListener(this);
	}

	/**
	 * Callback- Create and intialize parent control (table and actions).
	 */
	public void createPartControl(Composite parent) {
			
		Table table=new Table(parent,SWT.SINGLE | SWT.H_SCROLL | SWT.V_SCROLL|
				SWT.BORDER|SWT.FULL_SELECTION);
		contentProvider=new AnnoContentProvider();
		//******Create Gridlayout(table layout)******
		GridData grid=new GridData(GridData.FILL_BOTH);
		grid.grabExcessVerticalSpace=true;
		grid.horizontalSpan=4;
		table.setLayoutData(grid);
		table.setLinesVisible(true);
		table.setHeaderVisible(true);
		
		//********Columns*******		
		createColumns(table);
		
		//********Create Viewer*******
		viewer= new TableViewer(table);
		viewer.setContentProvider(contentProvider);
		viewer.setLabelProvider(new AnnoViewLabelProvider());
		viewer.setInput(getViewSite());
		
		//****Create and hook Actions****
		makeActions();
		hookContextMenu();
		hookDoubleClickAction();
		contributeToActionBars();
		
		//***********KeyListener**********
		table.addKeyListener(this);	
	}

	
	/**
	 * Adds columns to a given table.
	 * Columns: Reviewed, Description, Points, File
	 */
	private void createColumns(Table table) {
		
		TableColumn column;
		//Reviewed
		column=new TableColumn(table,SWT.CENTER,0);
		column.setText(columnLabels[0]);
		column.setWidth(stateWidth);
		column.setResizable(true);
		//Description
		column=new TableColumn(table,SWT.LEFT,1);
		column.setText(columnLabels[1]);
		column.setWidth(descWidth);
		column.setResizable(true);
		//Points
		column=new TableColumn(table,SWT.CENTER,2);
		column.setText(columnLabels[2]);
		column.setWidth(pointWidth);
		column.setResizable(true);
		//File
		column=new TableColumn(table,SWT.LEFT,3);
		column.setText(columnLabels[3]);
		column.setWidth(fileWidth);
		column.setResizable(true);
	}
	
	/**
	 * Set the annotation file to be viewed in the table.
	 * @param f- Annotation File
	 */
	public void setAnnotationFile(AnnotationFile f){
		annoFile=f;
		contentProvider.setAnnotationList(f.getAnnotationItems());
		viewer.refresh();
	}

	/**
	 * Returns the currently shown AnnotationFile
	 * @return AnnotationFile
	 */
	public AnnotationFile getAnnotationFile(){
		return annoFile;
	}
	
	/**
	 * Adds an Annotation Item to the currently shown AnnotationFile (List)
	 * @param a
	 * @return success
	 */
	public boolean addToAnnotationList(AnnoObject a){
		annoFile.addAnnotationItem(a);
		boolean ret=contentProvider.addToAnnotationList(a);
		viewer.refresh();
		return ret;
		
	}
	
	/**
	 * Hooks the context menu- Creates Menu and registrates it
	 * at the table viewer.
	 */
	private void hookContextMenu() {
		MenuManager menuMgr = new MenuManager("#PopupMenu",POPUPMENUID);
		menuMgr.setRemoveAllWhenShown(true);
		menuMgr.addMenuListener(new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {
				manager.add(actionReviewed);
				manager.add(doubleClickAction);
				manager.add(actionRemove);
				// Other plug-ins can contribute actions here
				manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
			}
		});
		Menu menu = menuMgr.createContextMenu(viewer.getControl());
		viewer.getControl().setMenu(menu);
		//register globaly
		getSite().registerContextMenu(menuMgr, viewer);
		
	}

	/**
	 * Adds actions to the ActionBar.
	 */
	private void contributeToActionBars() {
		IActionBars bars = getViewSite().getActionBars();
		IToolBarManager manager=bars.getToolBarManager();
		manager.add(actionReviewed);
		manager.add(doubleClickAction);
		manager.add(actionRemove);
		
	}

	/**
	 * Difines the Actions used:
	 * 	+remove
	 *  +review
	 *  +show source (double click!!)
	 */
	private void makeActions() {
		/*deletes a annotation item from the list */
		actionRemove = new Action() {
			public void run() {
				ISelection selection = viewer.getSelection();
				AnnoObject obj = (AnnoObject)((IStructuredSelection)selection).getFirstElement();
				if(annoFile==null){
					resourceMessage("AV_W_LISTUNINIT");
					return;
				}
				if(obj==null) return;
				if(!annoFile.removeAnnotationItem(obj)) resourceMessage("AV_W_NOTREMOVE");
				viewer.refresh();
				if(!annoFile.saveFile()){
					resourceMessage("AV_W_NOTSTORED");
				} 
			}
		};
		
		actionRemove.setText(AnnotationPlugin.getResourceString("AV_A_REMOVE"));
		actionRemove.setToolTipText(AnnotationPlugin.getResourceString("AV_TT_REMOVE"));
		actionRemove.setImageDescriptor(AnnotationPlugin.getDefault().getImageDescriptor(AnnotationPlugin.IMG_DEL));
		actionRemove.setAccelerator(SWT.DEL);
		
		/*sets the annotation item reviewed*/
		actionReviewed = new Action() {
			public void run() {
				ISelection selection = viewer.getSelection();
				AnnoObject obj = (AnnoObject)((IStructuredSelection)selection).getFirstElement();
				if(obj==null) return;
				if(!obj.isChecked()){
					obj.setChecked(true);
					viewer.refresh();
					if(!annoFile.saveFile()){
						resourceMessage("AV_W_NOTSTORED");
					}
				}
			}
		};
		actionReviewed.setText(AnnotationPlugin.getResourceString("AV_A_REVIEW"));
		actionReviewed.setToolTipText(AnnotationPlugin.getResourceString("AV_TT_REVIEW"));
		actionReviewed.setImageDescriptor(AnnotationPlugin.getDefault().getImageDescriptor(AnnotationPlugin.IMG_REVIEWED));
		actionReviewed.setAccelerator('R');
		
		/*show source of the annotation*/
		doubleClickAction = new Action() {
			public void run() {
				ISelection selection = viewer.getSelection();
				AnnoObject obj = (AnnoObject)((IStructuredSelection)selection).getFirstElement();
				if (obj==null) return;
				AbstractTextEditor editor=null;
				try {					
					
					IFile res=(IFile)ResourcesPlugin.getWorkspace().getRoot().
						findMember(annoFile.getProjectName()+obj.getFile());
					FileEditorInput fileinput=new FileEditorInput(res);
					editor=(AbstractTextEditor)PlatformUI.getWorkbench().getActiveWorkbenchWindow().
						getActivePage().openEditor(fileinput,JAVAEDITORID);
				} catch (Exception e) {
					resourceMessage("AV_W_FILENOTFOUND", annoFile.getProjectName()+obj.getFile());
					return;
				}
				if(editor!=null){
					//Restore the selection
					TextSelection sel=new TextSelection(obj.getOffset(),obj.getLength());
					editor.getSelectionProvider().setSelection(sel);
				}
			}
		};
		doubleClickAction.setText(AnnotationPlugin.getResourceString("AV_A_SHOW"));
		doubleClickAction.setToolTipText(AnnotationPlugin.getResourceString("AV_TT_SHOW"));
		doubleClickAction.setImageDescriptor(AnnotationPlugin.getDefault().getImageDescriptor(AnnotationPlugin.IMG_SHOW));
		doubleClickAction.setAccelerator('S');
		
	}

	/**
	 * Hooks the doubleclick action to the table viewer.
	 */
	private void hookDoubleClickAction() {
		viewer.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent event) {
				doubleClickAction.run();
			}
		});
	}
	
	/**
	 * Shows a message dialog.
	 * @param message
	 */
	private void showMessage(String message) {
		MessageDialog.openInformation(
			viewer.getControl().getShell(),
			AnnotationPlugin.getResourceString("AV_MESSAGE_TOP"),message);
	}
	
	/**
	 * Loads a message from resource bundle using key and
	 * shows it via showMessage.
	 * @param key
	 */
	private void resourceMessage(String key){
		showMessage(AnnotationPlugin.getResourceString(key));
	}

	/**
	 * Loads a message from resource bundle using key and
	 * shows it via showMessage adding the add string.
	 * @param key
	 * @param add
	 */
	private void resourceMessage(String key, String add){
		showMessage(AnnotationPlugin.getResourceString(key)+add);
	}
	
	/**
	 * Passing the focus request to the viewer's control.
	 */
	public void setFocus() {
		viewer.getControl().setFocus();
	}

	/**
	 * Resource Change Events call this methode
	 */
	public void resourceChanged(IResourceChangeEvent event) {
		if(event==null || annoFile==null) return;
		//the source file has changes- it is reloaded
		if(event.getType()==IResourceChangeEvent.POST_CHANGE){
			 
			IPath p=new Path(annoFile.getFileName());
			IFile f=ResourcesPlugin.getWorkspace().getRoot().getFile(p);
			
			//check if our resource fired change event 
			IResource res=null;
			IResourceDelta delta=event.getDelta().findMember(f.getFullPath());
			if(delta!=null) res=delta.getResource();
			
			if(res!=null){
				if(annoFile.refresh()) contentProvider.setAnnotationList(annoFile.getAnnotationItems());
				viewer.refresh();
			}	
		}
	}


	/**
	 * Handles user's keyboard input. 
	 */
	public void keyPressed(KeyEvent e) { 
		if(annoFile!=null && !viewer.getSelection().isEmpty()){
			//entf 
			if(e.keyCode==KEY_ENF) actionRemove.run();
			//r
			if(e.keyCode==KEY_R) actionReviewed.run();	
			//s
			if(e.keyCode==KEY_S) doubleClickAction.run();
			
		}
	}

	/**
	 * Not used.
	 * @param e
	 */
	public void keyReleased(KeyEvent e) {}
}